@@ -10,6 +10,7 @@ import (
 	"go/token"
 	"reflect"
 
+	"github.com/go-delve/delve/pkg/dwarf/godwarf"
 	"github.com/go-delve/delve/pkg/dwarf/op"
 	"github.com/go-delve/delve/pkg/dwarf/reader"
 	"github.com/go-delve/delve/pkg/goversion"
@@ -24,8 +25,15 @@ const (
 	// process dies because of a fatal runtime error.
 	FatalThrow = "runtime-fatal-throw"
 
-	unrecoveredPanicID = -1
-	fatalThrowID       = -2
+	// HardcodedBreakpoint is the name given to hardcoded breakpoints (for
+	// example: calls to runtime.Breakpoint)
+	HardcodedBreakpoint = "hardcoded-breakpoint"
+
+	unrecoveredPanicID    = -1
+	fatalThrowID          = -2
+	hardcodedBreakpointID = -3
+
+	NoLogicalID = -1000 // Logical breakpoint ID for breakpoints internal breakpoints.
 )
 
 // Breakpoint represents a physical breakpoint. Stores information on the break
@@ -40,7 +48,6 @@ type Breakpoint struct {
 	Addr         uint64 // Address breakpoint is set for.
 	OriginalData []byte // If software breakpoint, the data we replace with breakpoint instruction.
 	Name         string // User defined name of the breakpoint
-	LogicalID    int    // ID of the logical breakpoint that owns this physical breakpoint
 
 	WatchExpr     string
 	WatchType     WatchType
@@ -59,6 +66,7 @@ type Breakpoint struct {
 	Variables   []string // Variables to evaluate
 	LoadArgs    *LoadConfig
 	LoadLocals  *LoadConfig
+	UserData    interface{} // Any additional information about the breakpoint
 
 	// ReturnInfo describes how to collect return variables when this
 	// breakpoint is hit as a return breakpoint.
@@ -72,6 +80,8 @@ type Breaklet struct {
 	// stepping).
 	Kind BreakpointKind
 
+	LogicalID int // ID of the logical breakpoint that owns this physical breakpoint
+
 	// Cond: if not nil the breakpoint will be triggered only if evaluating Cond returns true
 	Cond ast.Expr
 
@@ -173,7 +183,16 @@ func (wtype WatchType) withSize(sz uint8) WatchType {
 var ErrHWBreakUnsupported = errors.New("hardware breakpoints not implemented")
 
 func (bp *Breakpoint) String() string {
-	return fmt.Sprintf("Breakpoint %d at %#v %s:%d", bp.LogicalID, bp.Addr, bp.File, bp.Line)
+	return fmt.Sprintf("Breakpoint %d at %#v %s:%d", bp.LogicalID(), bp.Addr, bp.File, bp.Line)
+}
+
+func (bp *Breakpoint) LogicalID() int {
+	for _, breaklet := range bp.Breaklets {
+		if breaklet.Kind == UserBreakpoint {
+			return breaklet.LogicalID
+		}
+	}
+	return NoLogicalID
 }
 
 // VerboseDescr returns a string describing parts of the breakpoint struct
@@ -358,6 +377,13 @@ func isPanicCall(frames []Stackframe) (bool, int) {
 }
 
 func isDeferReturnCall(frames []Stackframe, deferReturns []uint64) (bool, uint64) {
+	if len(frames) >= 2 && (len(deferReturns) > 0) {
+		// On Go 1.18 and later runtime.deferreturn doesn't use jmpdefer anymore,
+		// it's a normal function making normal calls to deferred functions.
+		if frames[1].Current.Fn != nil && frames[1].Current.Fn.Name == "runtime.deferreturn" {
+			return true, 0
+		}
+	}
 	if len(frames) >= 1 {
 		for _, pc := range deferReturns {
 			if frames[0].Ret == pc {
@@ -445,9 +471,6 @@ type BreakpointMap struct {
 	// WatchOutOfScope is the list of watchpoints that went out of scope during
 	// the last resume operation
 	WatchOutOfScope []*Breakpoint
-
-	breakpointIDCounter         int
-	internalBreakpointIDCounter int
 }
 
 // NewBreakpointMap creates a new BreakpointMap.
@@ -459,19 +482,60 @@ func NewBreakpointMap() BreakpointMap {
 
 // SetBreakpoint sets a breakpoint at addr, and stores it in the process wide
 // break point table.
-func (t *Target) SetBreakpoint(addr uint64, kind BreakpointKind, cond ast.Expr) (*Breakpoint, error) {
-	return t.setBreakpointInternal(addr, kind, 0, cond)
+func (t *Target) SetBreakpoint(logicalID int, addr uint64, kind BreakpointKind, cond ast.Expr) (*Breakpoint, error) {
+	return t.setBreakpointInternal(logicalID, addr, kind, 0, cond)
 }
 
+// SetEBPFTracepoint will attach a uprobe to the function
+// specified by 'fnName'.
 func (t *Target) SetEBPFTracepoint(fnName string) error {
+	// Not every OS/arch that we support has support for eBPF,
+	// so check early and return an error if this is called on an
+	// unsupported system.
 	if !t.proc.SupportsBPF() {
 		return errors.New("eBPF is not supported")
 	}
-	var args []ebpf.UProbeArgMap
-	fn, ok := t.BinInfo().LookupFunc[fnName]
-	if !ok {
-		return fmt.Errorf("could not find function %s", fnName)
+	fns, err := t.BinInfo().FindFunction(fnName)
+	if err != nil {
+		return err
 	}
+
+	// Get information on the Goroutine so we can tell the
+	// eBPF program where to find it in order to get the
+	// goroutine ID.
+	rdr := t.BinInfo().Images[0].DwarfReader()
+	rdr.SeekToTypeNamed("runtime.g")
+	typ, err := t.BinInfo().findType("runtime.g")
+	if err != nil {
+		return errors.New("could not find type for runtime.g")
+	}
+	var goidOffset int64
+	switch t := typ.(type) {
+	case *godwarf.StructType:
+		for _, field := range t.Field {
+			if field.Name == "goid" {
+				goidOffset = field.ByteOffset
+				break
+			}
+		}
+	}
+
+	for _, fn := range fns {
+		err := t.setEBPFTracepointOnFunc(fn, goidOffset)
+		if err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+func (t *Target) setEBPFTracepointOnFunc(fn *Function, goidOffset int64) error {
+	// Start putting together the argument map. This will tell the eBPF program
+	// all of the arguments we want to trace and how to find them.
+
+	// Start looping through each argument / return parameter for the function we
+	// are setting the uprobe on. Parse location information so that we can pass it
+	// along to the eBPF program.
 	dwarfTree, err := fn.cu.image.getDwarfTree(fn.offset)
 	if err != nil {
 		return err
@@ -482,16 +546,14 @@ func (t *Target) SetEBPFTracepoint(fnName string) error {
 	}
 	_, l, _ := t.BinInfo().PCToLine(fn.Entry)
 
+	var args []ebpf.UProbeArgMap
 	varEntries := reader.Variables(dwarfTree, fn.Entry, l, variablesFlags)
 	for _, entry := range varEntries {
-		isret, _ := entry.Val(dwarf.AttrVarParam).(bool)
-		if isret {
-			continue
-		}
 		_, dt, err := readVarEntry(entry.Tree, fn.cu.image)
 		if err != nil {
 			return err
 		}
+
 		offset, pieces, _, err := t.BinInfo().Location(entry, dwarf.AttrLocation, fn.Entry, op.DwarfRegisters{}, nil)
 		if err != nil {
 			return err
@@ -502,16 +564,28 @@ func (t *Target) SetEBPFTracepoint(fnName string) error {
 				paramPieces = append(paramPieces, int(piece.Val))
 			}
 		}
+		isret, _ := entry.Val(dwarf.AttrVarParam).(bool)
 		offset += int64(t.BinInfo().Arch.PtrSize())
-		args = append(args, ebpf.UProbeArgMap{Offset: offset, Size: dt.Size(), Kind: dt.Common().ReflectKind, Pieces: paramPieces, InReg: len(pieces) > 0})
+		args = append(args, ebpf.UProbeArgMap{
+			Offset: offset,
+			Size:   dt.Size(),
+			Kind:   dt.Common().ReflectKind,
+			Pieces: paramPieces,
+			InReg:  len(pieces) > 0,
+			Ret:    isret,
+		})
 	}
-	t.proc.SetUProbe(fnName, args)
+
+	//TODO(aarzilli): inlined calls?
+
+	// Finally, set the uprobe on the function.
+	t.proc.SetUProbe(fn.Name, goidOffset, args)
 	return nil
 }
 
 // SetWatchpoint sets a data breakpoint at addr and stores it in the
 // process wide break point table.
-func (t *Target) SetWatchpoint(scope *EvalScope, expr string, wtype WatchType, cond ast.Expr) (*Breakpoint, error) {
+func (t *Target) SetWatchpoint(logicalID int, scope *EvalScope, expr string, wtype WatchType, cond ast.Expr) (*Breakpoint, error) {
 	if (wtype&WatchWrite == 0) && (wtype&WatchRead == 0) {
 		return nil, errors.New("at least one of read and write must be set for watchpoint")
 	}
@@ -550,7 +624,7 @@ func (t *Target) SetWatchpoint(scope *EvalScope, expr string, wtype WatchType, c
 		return nil, errors.New("can not watch stack allocated variable for reads")
 	}
 
-	bp, err := t.setBreakpointInternal(xv.Addr, UserBreakpoint, wtype.withSize(uint8(sz)), cond)
+	bp, err := t.setBreakpointInternal(logicalID, xv.Addr, UserBreakpoint, wtype.withSize(uint8(sz)), cond)
 	if err != nil {
 		return bp, err
 	}
@@ -567,19 +641,32 @@ func (t *Target) SetWatchpoint(scope *EvalScope, expr string, wtype WatchType, c
 	return bp, nil
 }
 
-func (t *Target) setBreakpointInternal(addr uint64, kind BreakpointKind, wtype WatchType, cond ast.Expr) (*Breakpoint, error) {
+func (t *Target) setBreakpointInternal(logicalID int, addr uint64, kind BreakpointKind, wtype WatchType, cond ast.Expr) (*Breakpoint, error) {
 	if valid, err := t.Valid(); !valid {
-		return nil, err
+		recorded, _ := t.Recorded()
+		if !recorded {
+			return nil, err
+		}
 	}
 	bpmap := t.Breakpoints()
 	newBreaklet := &Breaklet{Kind: kind, Cond: cond}
 	if kind == UserBreakpoint {
 		newBreaklet.HitCount = map[int]uint64{}
+		newBreaklet.LogicalID = logicalID
 	}
 	if bp, ok := bpmap.M[addr]; ok {
 		if !bp.canOverlap(kind) {
 			return bp, BreakpointExistsError{bp.File, bp.Line, bp.Addr}
 		}
+		if kind == UserBreakpoint {
+			bp.Tracepoint = false
+			bp.TraceReturn = false
+			bp.Goroutine = false
+			bp.Stacktrace = 0
+			bp.Variables = nil
+			bp.LoadArgs = nil
+			bp.LoadLocals = nil
+		}
 		bp.Breaklets = append(bp.Breaklets, newBreaklet)
 		return bp, nil
 	}
@@ -620,14 +707,6 @@ func (t *Target) setBreakpointInternal(addr uint64, kind BreakpointKind, wtype W
 		return nil, err
 	}
 
-	if kind != UserBreakpoint {
-		bpmap.internalBreakpointIDCounter++
-		newBreakpoint.LogicalID = bpmap.internalBreakpointIDCounter
-	} else {
-		bpmap.breakpointIDCounter++
-		newBreakpoint.LogicalID = bpmap.breakpointIDCounter
-	}
-
 	newBreakpoint.Breaklets = append(newBreakpoint.Breaklets, newBreaklet)
 
 	bpmap.M[addr] = newBreakpoint
@@ -635,17 +714,6 @@ func (t *Target) setBreakpointInternal(addr uint64, kind BreakpointKind, wtype W
 	return newBreakpoint, nil
 }
 
-// SetBreakpointWithID creates a breakpoint at addr, with the specified logical ID.
-func (t *Target) SetBreakpointWithID(id int, addr uint64) (*Breakpoint, error) {
-	bpmap := t.Breakpoints()
-	bp, err := t.SetBreakpoint(addr, UserBreakpoint, nil)
-	if err == nil {
-		bp.LogicalID = id
-		bpmap.breakpointIDCounter--
-	}
-	return bp, err
-}
-
 // canOverlap returns true if a breakpoint of kind can be overlapped to the
 // already existing breaklets in bp.
 // At most one user breakpoint can be set but multiple internal breakpoints are allowed.
@@ -658,13 +726,16 @@ func (bp *Breakpoint) canOverlap(kind BreakpointKind) bool {
 }
 
 // ClearBreakpoint clears the breakpoint at addr.
-func (t *Target) ClearBreakpoint(addr uint64) (*Breakpoint, error) {
+func (t *Target) ClearBreakpoint(addr uint64) error {
 	if valid, err := t.Valid(); !valid {
-		return nil, err
+		recorded, _ := t.Recorded()
+		if !recorded {
+			return err
+		}
 	}
 	bp, ok := t.Breakpoints().M[addr]
 	if !ok {
-		return nil, NoBreakpointError{Addr: addr}
+		return NoBreakpointError{Addr: addr}
 	}
 
 	for i := range bp.Breaklets {
@@ -675,21 +746,21 @@ func (t *Target) ClearBreakpoint(addr uint64) (*Breakpoint, error) {
 
 	_, err := t.finishClearBreakpoint(bp)
 	if err != nil {
-		return nil, err
+		return err
 	}
 
 	if bp.WatchExpr != "" && bp.watchStackOff != 0 {
 		// stack watchpoint, must remove all its WatchOutOfScopeBreakpoints/StackResizeBreakpoints
 		err := t.clearStackWatchBreakpoints(bp)
 		if err != nil {
-			return bp, err
+			return err
 		}
 	}
 
-	return bp, nil
+	return nil
 }
 
-// ClearInternalBreakpoints removes all stepping breakpoints from the map,
+// ClearSteppingBreakpoints removes all stepping breakpoints from the map,
 // calling clearBreakpoint on each one.
 func (t *Target) ClearSteppingBreakpoints() error {
 	bpmap := t.Breakpoints()
@@ -838,7 +909,7 @@ func (rbpi *returnBreakpointInfo) Collect(t *Target, thread Thread) []*Variable
 		return returnInfoError("could not read function entry", err, thread.ProcessMemory())
 	}
 
-	vars, err := scope.Locals()
+	vars, err := scope.Locals(0)
 	if err != nil {
 		return returnInfoError("could not evaluate return variables", err, thread.ProcessMemory())
 	}
